part of enhanced_ddp;

typedef UpdateListener = void Function(
  String collection,
  String operation,
  String id,
  Map<String, dynamic> doc,
);

Tuple2<String, Map<String, dynamic>> _parseUpdate(Map<String, dynamic> update) {
  if (update.containsKey('id')) {
    final id = update['id'];
    if (id.runtimeType == String) {
      if (update.containsKey('fields')) {
        final updates = update['fields'];
        if (updates is Map) {
          return Tuple2(id, updates);
        }
      }
      return Tuple2(id, null);
    }
  }
  return Tuple2('', null);
}

abstract class Collection {
  Map<String, dynamic> findOne(String id);

  Map<String, Map<String, dynamic>> findAll();

  void addUpdateListener(UpdateListener listener);

  void removeUpdateListeners();

  void _added(Map<String, dynamic> msg);

  void _changed(Map<String, dynamic> msg);

  void _removed(Map<String, dynamic> msg);

  void _addedBefore(Map<String, dynamic> msg);

  void _movedBefore(Map<String, dynamic> msg);

  void _init();

  void _reset();

  factory Collection.mock() => _MockCache();

  factory Collection.key(String name) => KeyCache(name, {}, []);
}

class KeyCache implements Collection {
  String name;
  // ignore: prefer_final_fields
  Map<String, Map<String, dynamic>> _items;
  // ignore: prefer_final_fields
  List<UpdateListener> _listeners;

  KeyCache(name, _items, _listeners);

  void _notify(String operation, String id, Map<String, dynamic> doc) {
    _listeners.forEach((listener) {
      listener(name, operation, id, doc);
    });
  }

  @override
  void _added(Map<String, dynamic> msg) {
    final pair = _parseUpdate(msg);
    if (pair.item2 != null) {
      _items[pair.item1] = pair.item2;
      _notify('create', pair.item1, pair.item2);
    }
  }

  @override
  void _addedBefore(Map<String, dynamic> msg) {}

  @override
  void _changed(Map<String, dynamic> msg) {
    final pair = _parseUpdate(msg);
    if (pair.item2 != null) {
      if (_items.containsKey(pair.item1)) {
        final item = _items[pair.item1];
        pair.item2.forEach((key, value) => item[key] = value);
        _items[pair.item1] = item;
        _notify('update', pair.item1, item);
      }
    }
  }

  @override
  void _init() {}

  @override
  void _movedBefore(Map<String, dynamic> msg) {}

  @override
  void _removed(Map<String, dynamic> msg) {
    final pair = _parseUpdate(msg);
    if (pair.item1.isNotEmpty) {
      _items.remove(pair.item1);
      _notify('remove', pair.item1, null);
    }
  }

  @override
  void _reset() {
    _notify('reset', '', null);
  }

  @override
  void addUpdateListener(UpdateListener listener) {
    _listeners.add(listener);
  }

  @override
  void removeUpdateListeners() {
    _listeners.clear();
  }

  @override
  Map<String, Map<String, dynamic>> findAll() => _items;

  @override
  Map<String, dynamic> findOne(String id) => _items[id];
}

class _MockCache implements Collection {
  @override
  void _added(Map<String, dynamic> msg) {}

  @override
  void _addedBefore(Map<String, dynamic> msg) {}

  @override
  void _changed(Map<String, dynamic> msg) {}

  @override
  void _init() {}

  @override
  void _movedBefore(Map<String, dynamic> msg) {}

  @override
  void _removed(Map<String, dynamic> msg) {}

  @override
  void _reset() {}

  @override
  void addUpdateListener(UpdateListener listener) {}

  @override
  void removeUpdateListeners() {}

  @override
  Map<String, Map<String, dynamic>> findAll() => {};

  @override
  Map<String, dynamic> findOne(String id) => null;
}
